local _, ns = ...
local WK = ns.WhoKnows
WK.PlayerStats = {}
--------------------------------------------------------------------------
local print = WK.print
local debugPrint = WK.debugPrint
WK.DebugTable.skillsRecorded = {}
WK.DebugTable.dupeSpells = {}
--------------------------------------------------------------------------

WK.ItemMessages = {
	NONE_CAN_LEARN              = "|cffC79C6ENone of your toons can learn this skill.|r",
	BOP                         = "|cffffe700Unlearnable: BOP item. Only checked current toon.",
	KNOWN                       = " : |cff008500known|r",
	NOT_KNOWN                   = " : |cffC41F3BNOT known|r",
	INCOMPLETE                  = " : |cffffe700Incomplete records. Please load toon and scan professions.|r",
	NEEDS_PREREQ                = " : |cffffe700Prerequisites need to be acquired or met first..|r",
	SKILL_LVL_UNK               = "\n     Skill level unknown",
	SKILL_LVL_TOO_LOW           = "\n     |cffffe700Skill too low to learn (currently: %s needs: %s)|r", --curToonSkillLevel,reqSkillLvl
	LVL_UNK                     = "\n     Toon level unknown",
	LVL_TOO_LOW                 = "\n     |cffffe700Player's level too low. (currently %s needs level %s)|r", -- curToonLevel,reqToonLvl
	UNK_IF_CAN_LEARN            = "\n     Unknown if can learn",
	LEARN_WITH_RESTRICTIONS     = "\n     |cffC41F3BWarning:|cffffe700Class restrictions: May be able to learn but BOP item cannot be equipped.|r",
	UNABLE_TO_LEARN             = "\n     |cffff8000Unable to learn at this time.|r",
	ABLE_TO_LEARN               = "\n     |cff1eff00Able to learn|r",
	NEEDS_ENG_SPEC              = "\n     |cffffe700This toon will need to learn the %s engineering spec to learn this.|r", -- reqEngiSpec
	WRONG_ENG_SPEC              = "\n     |cffffe700This toon has the wrong Engineering spec. Requires %s Engineer.|r", -- reqEngiSpec
	REQ_PREV_RANK				= "\n     |cffff8000Need to learn a lower rank of this" -- reqPreviousRank
}

WK.ItemStatus = {
	UNABLE_TO_LEARN         = 0,
	ABLE_TO_LEARN           = 1,
	UNK_IF_CAN_LEARN        = 2,
	LEARN_WITH_RESTRICTIONS = 3,
	INCOMPLETE_INFO         = 4,
	NEEDS_PREREQ            = 5,
	ALREADY_KNOWN           = 6,
}

-------------------------------------------------------
function WK.PlayerStats:GetToonLearnedStatus(item)
	if ((item.itemClass == "Recipe") and WhoKnows_Data.EnableRecipeCheck) then
		local findEligibleToons = WK.PlayerStats:RecipeKnownBy(item)
		local toonSearchResults = WK.PlayerStats:DetermineLearnability(findEligibleToons, item)
		return toonSearchResults
	elseif (item.isBlueprint and WhoKnows_Data.GarrisonBlueprintCheck) then
		return WK.PlayerStats:SpecialMiscRecipe(item)
	end
end

-------------------------------------------------------
function WK.PlayerStats:DetermineLearnability(toons, item)
	local realm = WK.currentPlayer.realm
	local faction = WK.currentPlayer.faction
	local toonSearchResults = {} -- #table holds toons

	if (WK:tlength(toons) == 0) then
		toonSearchResults["*"] = {}
		if item.isBOP then
			toonSearchResults["*"]["msg"] = WK.ItemMessages.BOP
		else
			toonSearchResults["*"]["msg"] = WK.ItemMessages.NONE_CAN_LEARN
		end
		toonSearchResults["*"]["status"] = 0
		return toonSearchResults
	end

	for toon, isKnown in pairs(toons) do
		toonSearchResults[toon] = {}
		if isKnown == true then
			toonSearchResults[toon]["msg"] = WK.ItemMessages.KNOWN
			toonSearchResults[toon]["status"] = 6
		elseif isKnown == "incomplete" then
			toonSearchResults[toon]["msg"] = WK.ItemMessages.INCOMPLETE
			toonSearchResults[toon]["status"] = 4
		elseif item.needsPrereq then
			toonSearchResults[toon]["msg"] = WK.ItemMessages.NEEDS_PREREQ
			toonSearchResults[toon]["status"] = 5
		elseif item.isNotLearnableYet and item.isBOP then
			toonSearchResults[toon]["msg"] = WK.ItemMessages.NEEDS_PREREQ
			toonSearchResults[toon]["status"] = 5
		elseif item.reqPreviousRank then
			toonSearchResults[toon]["msg"] = WK.ItemMessages.REQ_PREV_RANK
			toonSearchResults[toon]["status"] = 5
		else
			local learnStatus = WK.ItemStatus.ABLE_TO_LEARN
			local msg = ""
			local toonSkillLevel = (WhoKnows_Data.Skills[realm][faction][toon]["skillLvls"][item.profession] or -1)
			if item.reqClass and #item.reqClass > 0 then
				local isClass = false
				for _, v in pairs(item.reqClass) do
					isClass = (v == WhoKnows_Data.Skills[realm][faction][toon]["class"])
					if isClass == true then break end
				end
				if isClass == false then
					learnStatus = WK.ItemStatus.LEARN_WITH_RESTRICTIONS
				end
			end
			if ((learnStatus < WK.ItemStatus.LEARN_WITH_RESTRICTIONS) and item.reqSkillLvl > toonSkillLevel) then
				if toonSkillLevel == -1 then
					msg = msg .. "\n     Skill level unknown"
					learnStatus = WK.ItemStatus.UNK_IF_CAN_LEARN
				else
					msg = msg .. string.format(WK.ItemMessages.SKILL_LVL_TOO_LOW, toonSkillLevel, item.reqSkillLvl)
					learnStatus = WK.ItemStatus.UNABLE_TO_LEARN
				end
			end
			local toonLevel = (WhoKnows_Data.Skills[realm][faction][toon]["level"] or -1)
			if ((learnStatus < WK.ItemStatus.LEARN_WITH_RESTRICTIONS) and item.reqToonLvl > toonLevel) then
				if toonLevel == -1 then
					msg = msg .. "\n     Toon level unknown"
					learnStatus = WK.ItemStatus.UNK_IF_CAN_LEARN
				else
					msg = msg .. string.format(WK.ItemMessages.LVL_TOO_LOW, toonLevel , item.reqToonLvl)
					learnStatus = WK.ItemStatus.UNABLE_TO_LEARN
				end
			end
			if (item.reqEngiSpec and WhoKnows_Data.Skills[realm][faction][toon]["skillLvls"]["EngineeringSpec"]) then
				if WhoKnows_Data.Skills[realm][faction][toon]["skillLvls"]["EngineeringSpec"] == "Undeclared" then
					msg = msg .. string.format(WK.ItemMessages.NEEDS_ENG_SPEC, item.reqEngiSpec )
					learnStatus = WK.ItemStatus.UNABLE_TO_LEARN
				elseif item.reqEngiSpec ~= WhoKnows_Data.Skills[realm][faction][toon]["skillLvls"]["EngineeringSpec"] then
					msg = msg .. string.format(WK.ItemMessages.WRONG_ENG_SPEC, item.reqEngiSpec )
					learnStatus = WK.ItemStatus.UNABLE_TO_LEARN
				end
			end

			if learnStatus == WK.ItemStatus.UNABLE_TO_LEARN then
				msg = msg .. WK.ItemMessages.UNABLE_TO_LEARN
			elseif learnStatus == WK.ItemStatus.ABLE_TO_LEARN then
				msg = msg .. WK.ItemMessages.ABLE_TO_LEARN
			elseif learnStatus == WK.ItemMessages.REQ_PREV_RANK then
				msg = msg .. WK.ItemMessages.REQ_PREV_RANK
			elseif learnStatus == WK.ItemStatus.UNK_IF_CAN_LEARN then
				msg = msg .. WK.ItemMessages.UNK_IF_CAN_LEARN
			elseif learnStatus == WK.ItemStatus.LEARN_WITH_RESTRICTIONS then
				msg = WK.ItemMessages.LEARN_WITH_RESTRICTIONS
			end

			toonSearchResults[toon]["msg"] = WK.ItemMessages.NOT_KNOWN .. msg
			toonSearchResults[toon]["status"] = learnStatus
		end
	end
	return toonSearchResults
end

-------------------------------------------------------
function WK.PlayerStats:SpecialMiscRecipe(item)
	local resultsTable = {}
	resultsTable[WK.currentPlayer.name] = {}
	if item.isNotLearnableYet then
		resultsTable[WK.currentPlayer.name]["msg"] = WK.ItemMessages.UNABLE_TO_LEARN
		resultsTable[WK.currentPlayer.name]["status"] = WK.ItemStatus.NEEDS_PREREQ
	elseif item.isKnownBySelf then
		resultsTable[WK.currentPlayer.name]["msg"] = WK.ItemMessages.KNOWN
		resultsTable[WK.currentPlayer.name]["status"] = WK.ItemStatus.ALREADY_KNOWN
	else
		resultsTable[WK.currentPlayer.name]["msg"] = WK.ItemMessages.ABLE_TO_LEARN
		resultsTable[WK.currentPlayer.name]["status"] = WK.ItemStatus.ABLE_TO_LEARN
	end
	return resultsTable
end

-------------------------------------------------------
--- Scans all the player's toons in current realm to see if they know a given recipe
-- @param #WhoKnows.items.item A WoW item
-- @return #table A table of toons that know the item.
function WK.PlayerStats:RecipeKnownBy(item)
	local realm = WK.currentPlayer.realm
	local faction = WK.currentPlayer.faction
	local player = WK.currentPlayer.player
	local resultsTable = {}

	--BOP? Then only check current toon
	if item.isBOP then
		if player["skillsKnown"][item.profession] then
			resultsTable[WK.currentPlayer.name] = item.isKnownBySelf or false
		end
		return resultsTable
	end
	-- parse toons looking for item
	for searchPlayer in pairs(WhoKnows_Data.Skills[realm][faction]) do
		local currentPlayer = WhoKnows_Data.Skills[realm][faction][searchPlayer]
		if not currentPlayer["excludeFromScan"] then
			if currentPlayer["skillsKnown"][item.profession] then
				if currentPlayer == player and item.isKnownBySelf then
					resultsTable[searchPlayer] = true
				elseif (WK:tlength(currentPlayer["skillsKnown"][item.profession]) == 0) and currentPlayer["skillLvls"][item.profession] > 0 then
					resultsTable[searchPlayer] = "incomplete"
				else
					local index = currentPlayer["skillsKnown"][item.profession][item.spellTaughtId]
					--debugPrint(string.format("Looking to see if %s knows %s : %s - SpellTaughtId: %s",searchPlayer or "noName", item.profession or "noProf", item.itemName or "NoName", item.spellTaughtId or "noSpellId"))
					resultsTable[searchPlayer] = (index ~= nil)
				end
			end
		end
	end
	return resultsTable
end

-------------------------------------------------------
--- Gets all of the current player's professions and skill levels and saves to WhoKnows_Data
function WK.PlayerStats:GetPlayerSkillLvls()
	local realm = WK.currentPlayer.realm
	local faction = WK.currentPlayer.faction
	local player = WK.currentPlayer.player

	assert(realm ~= nil, "Error: WK.currentPlayer.realm is nil. Shouldn't be.")
	assert(faction ~= nil, "Error: WK.currentPlayer.faction is nil. Shouldn't be.")
	assert(player ~= nil, "Error: WK.currentPlayer.player is nil. Shouldn't be.")

	debugPrint("Getting Profesions")
	local profs = { GetProfessions() }
	debugPrint("found ".. #profs .. " to lookup")
	for i,v in pairs(profs) do
		local profName, _, profSkillLevel = GetProfessionInfo(v)
		debugPrint("Found "..profName)
		if (profName ~= "Fishing" and profName ~= "Archaeology") then
			player["skillLvls"][profName] = profSkillLevel
			if not player["skillsKnown"][profName] then
				player["skillsKnown"][profName] = {}
				debugPrint("adding "..profName)
			end
		end
	end
	--if engineering, check for Gnomish or Goblin spec
	if player["skillLvls"]["Engineering"] then
		if IsSpellKnown(20219) then -- Gnomish
		player["skillLvls"]["EngineeringSpec"] = "Gnomish"
		elseif IsSpellKnown(20222) then -- Goblin
		player["skillLvls"]["EngineeringSpec"] = "Goblin"
		else
			player["skillLvls"]["EngineeringSpec"] = "Undeclared"
		end
	end
end

-------------------------------------------------------
--- Scans the opened profession and verifies that we have all learned skills
--
function WK.PlayerStats:GetTradeSkills()
	if C_TradeSkillUI.IsTradeSkillLinked()
		or C_TradeSkillUI.IsTradeSkillGuild()
		or not C_TradeSkillUI.IsTradeSkillReady()
	then return end
	WK.WKEvFrame:UnregisterEvent("TRADE_SKILL_UPDATE")

	--clear out item cache
	--TODO: why? table.wipe(WK.items)
	local tradeskillName = select(2,C_TradeSkillUI.GetTradeSkillLine())
	if tradeskillName == "Runeforging" or tradeskillName == "Archaeology" then return end
	--debugPrint("tradeskillName = ".. tradeskillName)

	local player = WK.currentPlayer.player
	--save link to these tradeskills
	if (tradeskillName ~= "First Aid" and not player[tradeskillName .. "Link"]) then
		player[tradeskillName .. "Link"] = C_TradeSkillUI.GetTradeSkillListLink()
	end

	--needed?
	if not player["skillsKnown"][tradeskillName] then
		debugPrint("WARNING: Guess I needed to lookup player[skillsKnown][tradeskillName] in WK.PlayerStats:GetTradeSkills()")
		WK.PlayerStats:GetPlayerSkillLvls()
	end

	local recipeIDs = C_TradeSkillUI.GetFilteredRecipeIDs()
	--sanity check  - Also might be true if player opens up TradeSkill frame immediately after logging in.
	-- So to be sure, wait x seconds and try again.
	if #recipeIDs == 0 then
		debugPrint("WARNING: C_TradeSkillUI.GetFilteredRecipeIDs = 0??")
		if not WK.rescanned then
			debugPrint("Trying again in 5 seconds")
			C_Timer.After(5, WK.PlayerStats:GetTradeSkills())
			WK.rescanned = true
		end
		return
	end

	local skillCount = 0
	local skillAddedCounter = 0
	local recipeInfo = {}
	for i = 1, #recipeIDs do
		recipeInfo = C_TradeSkillUI.GetRecipeInfo(recipeIDs[i], recipeInfo)
		if recipeInfo.learned == true then
			local skillToSearch = recipeInfo.name:lower()
			skillCount = skillCount + 1
			local index = player["skillsKnown"][tradeskillName][recipeIDs[i]]
			local itemNotes = string.format("%s > %s > %s", parentCategory or "", skillCategory or "", skillToSearch)
			if index == nil then -- add it to list
				player["skillsKnown"][tradeskillName][recipeIDs[i]] = itemNotes
				skillAddedCounter = skillAddedCounter + 1
			end
		end
	end
	debugPrint("WK sees " .. skillCount .. " skills known...")
	debugPrint(" and has " .. WK:tlength(player["skillsKnown"][tradeskillName]) .. " recorded for " .. tradeskillName)
	if skillAddedCounter and skillAddedCounter > 0 then print(tradeskillName..": Added " .. skillAddedCounter .. " skills to WhoKnows.") end
end

function WK.PlayerStats:UpdateLearnedSkill(skill)
	local _, link, _, _, _, _, tradeskillName = GetItemInfo(skill)
	local itemId = WK.GetItemIdFromHyperlink(link)
	if not skill or not tradeskillName then
		debugPrint("WK.PlayerStats:UpdateLearnedSkill(skill): skill: " .. skill or "nil" .. " prof: " .. tradeskillName or "nil")
		debugPrint("Why is one of these nil? aborting")
		return
	end
	skill = skill:lower()
	debugPrint("running PlayerStats:UpdateLearnedSkill with new skill: " .. skill .. "\n and profession: " .. tradeskillName)
	local player = WK.currentPlayer.player
	local skillCategory, parentCategory = WK.GetSpellCategoryFromID(itemId)
	debugPrint("WK.PlayerStats:UpdateLearnedSkill(skill): Have itemid: " .. (itemId or "nil"))
	local index = player["skillsKnown"][tradeskillName][skill]
	if index == nil then -- add it to list
		player["skillsKnown"][tradeskillName][skill] = skillCategory
		debugPrint(string.format("Added: %s - %s, parentcategory: %s", skill, skillCategory or "<none>", parentCategory or "<none>"))
	end
end
